import { FieldNode, FragmentDefinitionNode, GraphQLAbstractType, GraphQLField, GraphQLObjectType, GraphQLResolveInfo, GraphQLSchema, InlineFragmentNode } from "graphql";
declare type FieldInfo = {
    fieldNode: FieldNode;
    field: GraphQLField<any, any>;
    objectType: GraphQLObjectType;
    schema: GraphQLSchema;
};
declare type AbstractFieldInfo = {
    fragmentNode: InlineFragmentNode | FragmentDefinitionNode;
    abstractType: GraphQLAbstractType;
    concreteType: GraphQLObjectType;
    field: GraphQLField<any, any>;
    schema: GraphQLSchema;
};
export declare type GetFieldNameFunc = (info: FieldInfo) => string | string[] | null;
export declare type GetAbstractTypeFieldMaskPathsFunc = (info: AbstractFieldInfo, getFieldMaskPaths: () => string[]) => string[];
export declare type FieldMaskPathsFromResolveInfoOptions = {
    /**
     * Get field name in field mask path.
     * If return `null`, this field is not included in field mask paths.
     */
    getFieldName?: GetFieldNameFunc;
    /**
     * Determine field mask paths in abstract type.
     * By default, it returns only the fields of the abstract type itself and ignores fragments of concrete types,
     * but you can change this behavior by defining `getAbstractTypeFieldMaskPaths`.
     */
    getAbstractTypeFieldMaskPaths?: GetAbstractTypeFieldMaskPathsFunc;
};
/**
 * Create field mask paths from `GraphQLResolveInfo`.
 *
 * @param typename Name of GraphQL object type
 * @param info
 * @param opts
 * @returns field mask paths
 *
 * @example
 * ```ts
 * import { FieldMask } from "google-protobuf/google/protobuf/field_mask_pb";
 * import { fieldMaskPathsFromResolveInfo } from "graphql-field-mask";
 *
 * const queryType = new GraphQLObjectType({
 *   name: "Query",
 *   fields: {
 *     viewer: {
 *       type: User,
 *       resolve(_source, _args, ctx, info) {
 *         const paths = fieldMaskPathsFromResolveInfo("User", info);
 *         const mask = new FieldMask().setPathsList(paths);
 *
 *         // ...
 *       }
 *     }
 *   }
 * })
 * ```
 *
 * ### Convert to snake case
 * ```ts
 * import { snakeCase } from "change-case";
 * import { fieldMaskPathsFromResolveInfo, GetFieldNameFunc } from "graphql-field-mask";
 *
 * const getFieldName: GetFieldNameFunc = ({ field }) => snakeCase(field.name);
 *
 * resolve(_source, _args, ctx, info) {
 *   const paths = fieldMaskPathsFromResolveInfo("User", info, { getFieldName });
 *   const mask = new FieldMask().setPathsList(paths);
 *
 *   // ...
 * }
 * ```
 *
 * ### With custom scalar
 * ```ts
 * import { getNamedType, isScalarType } from "graphql";
 * import { fieldMaskPathsFromResolveInfo, GetFieldNameFunc } from "graphql-field-mask";
 *
 * const getFieldName: GetFieldNameFunc = ({ field }) => {
 *   const fieldType = getNamedType(field.type);
 *   if (isScalarType(fieldType)) {
 *     switch (fieldType.name) {
 *     case 'Date':
 *       return ['year', 'month', 'day'].map(c => `${fieldName}.${c}`);
 *     // ...
 *     }
 *   }
 *   return field.name
 * };
 *
 * resolve(_source, _args, ctx, info) {
 *   const paths = fieldMaskPathsFromResolveInfo("User", info, { getCustomScalarFieldMaskPaths });
 *   const mask = new FieldMask().setPathsList(paths);
 *
 *   // ...
 * }
 * ```
 *
 * ### With [ProtoNexus](https://github.com/proto-graphql/proto-nexus)
 * ```ts
 * import { ProtobufFieldExtensions, ProtobufMessageExtensions, ProtobufOneofExtensions } from "proto-nexus";
 * import { fieldMaskPathsFromResolveInfo, GetFieldNameFunc } from "graphql-field-mask";
 *
 * const getFieldName: GetFieldNameFunc = ({ field }) => {
 *   const ext = (field.extensions ?? {}) as Partial<ProtobufFieldExtensions>;
 *   return ext.protobufField?.name  ?? null;
 * };
 *
 * const getAbstractTypeFieldMaskPaths: GetAbstractTypeFieldMaskPathsFunc = (info, getFieldMaskPaths) => {
 *   const oneofExt = (info.abstractType.extensions ?? {}) as Partial<ProtobufOneofExtensions>;
 *   const objExt = (info.concreteType.extensions ?? {}) as Partial<ProtobufMessageExtensions>;
 *   const prefix = (oneofExt.protobufOneof.fields ?? []).find(f => f.type === objExt.protobufMessage?.fullName)?.name;
 *   return prefix ? getFieldMaskPaths().map(p => `${prefix}.${p}`) : []
 * }
 *
 * resolve(_source, _args, ctx, info) {
 *   const paths = fieldMaskPathsFromResolveInfo("User", info, { getFieldName, getAbstractTypeFieldMaskPaths });
 *   const mask = new FieldMask().setPathsList(paths);
 *   // ...
 * }
 * ```
 */
export declare function fieldMaskPathsFromResolveInfo(typename: string, info: GraphQLResolveInfo, opts?: FieldMaskPathsFromResolveInfoOptions): string[];
export {};
//# sourceMappingURL=fieldMaskPathsFromResolveInfo.d.ts.map